from datasets import load_dataset, load_from_disk, Dataset
import torch
from argparse import ArgumentParser, ArgumentTypeError
import torch.distributed as dist
from tqdm import tqdm
import joblib
import itertools
import shutil
import os
import subprocess
import open_clip
from transformers import CLIPModel, CLIPProcessor
from torch.distributed import destroy_process_group
import datetime
from pathlib import Path
import huggingface_hub
from lens.utils import ddp_setup, create_sampler, create_dataloader, is_main_process, default_device

DICT_CLIP_NAME = {'openai/clip-vit-large-patch14': 'ViT_L_14',
                  'openai/clip-vit-base-patch16': 'ViT_B_16',
                  'openai/clip-vit-base-patch32': 'ViT_B_32',
                  'hf-hub:laion/CLIP-ViT-H-14-laion2B-s32B-b79K': 'LAION_ViT_H_14_2B',
                  'hf-hub:laion/CLIP-ViT-bigG-14-laion2B-39B-b160k': 'LAION_ViT_bigG_14_2B'}


def flatten(l):
    return [item for sublist in l for item in sublist]


class OpenaiclipDataset:
    def __init__(self, ds, transform, image_col, id_col):
        self.ds = ds
        self.image_col
        self.id_col
        self.transform = transform

    def _load_transform(self):
        return CLIPProcessor.from_pretrained(self.vision_encoder_name)

    def get_model(self):
        return CLIPModel.from_pretrained(self.vision_encoder_name)

    def __getitem__(self, idx):
        row = self.ds[idx]
        image = row[self.image_col].convert('RGB')
        id = row[self.id_col]

        image = self.transform(images=image, return_tensors="pt")[
            'pixel_values'].squeeze(0)
        return {'image': image,
                'id': torch.tensor(id, dtype=torch.int32)}

    def __len__(self):
        return len(self.ds)


class LaionclipDataset:
    def __init__(self, ds, transform, image_col, id_col):
        self.ds = ds
        self.image_col = image_col
        self.id_col = id_col
        self.transform = transform

    def __getitem__(self, idx):
        row = self.ds[idx]
        image = row[self.image_col].convert('RGB')
        id = row[self.id_col]

        image = self.transform(image)
        return {'image': image,
                'id': torch.tensor(id, dtype=torch.int32)}

    def __len__(self):
        return len(self.ds)


class Tagger:
    def __init__(self,
                 vision_encoder_name="hf-hub:laion/CLIP-ViT-H-14-laion2B-s32B-b79K",
                 dataset_vocab="Multimodal-Fatima/descriptors-text-davinci-003",
                 split_vocab="full",
                 zeroshot_weights="zw_attributes_laion_ViT_H_14_2B_descriptors_text_davinci_003_full.pt",
                 device=default_device):

        self.device = device

        if 'openai' in vision_encoder_name:
            self.transform = CLIPProcessor.from_pretrained(vision_encoder_name)
            self.model = CLIPModel.from_pretrained(
                vision_encoder_name).to(self.device)
            self.dataset_class = OpenaiclipDataset

        elif 'laion' in vision_encoder_name:
            model_and_transform = open_clip.create_model_and_transforms(
                vision_encoder_name)
            self.model = model_and_transform[0].to(self.device)
            self.transform = model_and_transform[2]
            self.dataset_class = LaionclipDataset

        self.vocabulary = flatten(load_dataset(
            dataset_vocab, split=split_vocab)['prompt_descriptions'])

        huggingface_hub.hf_hub_download(repo_id="Multimodal-Fatima/AttributesCLIP", filename=zeroshot_weights, local_dir=str(Path(__file__).parent / "weights"))
        self.zeroshot_weights = torch.load(
            str(Path(__file__).parent / f"weights/{zeroshot_weights}"), map_location=self.device).float()

    def get_tags(self,
                 ds,
                 image_col="image",
                 id_col="id",
                 top_k=5,
                 contrastive_th=0.2,
                 batch_size=8,
                 num_workers=0,
                 distributed_sampling=False):

        dataset = self.dataset_class(ds, self.transform, image_col, id_col)

        sampler = create_sampler(dataset, distributed=distributed_sampling)
        dataloader = create_dataloader(
            dataset, sampler, batch_size=batch_size, num_workers=num_workers)

        result = []

        for batch in tqdm(dataloader):
            with torch.no_grad():
                image = batch['image'].to(self.device)
                image_id = batch['id']
                try:
                    image_features = self.model.encode_image(image)
                except:
                    image_features = self.model.get_image_features(
                        pixel_values=image)

                image_features /= image_features.norm(dim=-1, keepdim=True)
                text_scores = image_features @ self.zeroshot_weights
                del image_features
                torch.cuda.empty_cache()
                top_scores, top_indexes = text_scores.float().cpu().topk(k=top_k, dim=-1)
                for scores, indexes, _id in zip(top_scores, top_indexes, image_id):
                    filter_indexes = indexes[scores >= contrastive_th]
                    if len(filter_indexes) > 0:
                        top_k_tags = [self.vocabulary[index]
                                      for index in filter_indexes]
                    else:
                        top_k_tags = []
                    result.append({id_col: _id.item(), "tags": top_k_tags})

        result_ds = Dataset.from_dict(
            {key: [d[key] for d in result] for key in result[0]})
        return result_ds


def parse_args():
    args = ArgumentParser(description="")
    # Dataset variables
    args.add_argument('--dataset_name',
                      required=True,
                      type=str,
                      default='Multimodal-Fatima/OxfordPets')

    args.add_argument('--dataset_split',
                      required=True,
                      type=str,
                      default='test')
    args.add_argument('--id_col', type=str, default='id')
    args.add_argument('--image_col', type=str, default='image')
    args.add_argument('--save_disk', action="store_true")

    # Vocab variables
    args.add_argument('--dataset_vocab', type=str,
                      default='Multimodal-Fatima/descriptors-text-davinci-003')
    args.add_argument('--split_vocab', type=str, default='full')

    # Model variables
    args.add_argument('--vision_encoder_name',
                      required=True,
                      type=str,
                      default='openai/clip-vit-large-patch14')

    args.add_argument('--zeroshot_weights',
                      required=True,
                      type=str,
                      default='weights/attributes_clip/zw_attributes_laion_ViT_H_14_2B_descriptors_text_davinci_003_full.pt')

    # Parameters
    args.add_argument('--batch_size',
                      type=int,
                      default=8)

    args.add_argument('--num_workers',
                      type=int,
                      default=0)

    args.add_argument('--top_k',
                      type=int,
                      default=5)
    args.add_argument('--contrastive_th',
                      type=float,
                      default=0.2)
    args = args.parse_args()
    return args


def main():
    num_tasks = int(os.environ["WORLD_SIZE"])
    global_rank = int(os.environ["RANK"])
    local_rank = int(os.environ["LOCAL_RANK"])

    args = parse_args()
    ds = load_dataset(args.dataset_name, split=args.dataset_split)
    tagger = Tagger(vision_encoder_name=args.vision_encoder_name,
                    dataset_vocab=args.dataset_vocab,
                    split_vocab=args.split_vocab,
                    zeroshot_weights=args.zeroshot_weights,
                    device=local_rank)
    result_ds = tagger.get_tags(ds,
                                image_col=args.image_col,
                                id_col=args.id_col,
                                top_k=args.top_k,
                                contrastive_th=args.contrastive_th,
                                batch_size=args.batch_size,
                                num_workers=args.num_workers,
                                distributed_sampling=True)

    # Save in pickle all embeddings
    os.makedirs(Path(__file__).parent /
                "attributes_clip_tmp_folder/", exist_ok=True)
    result = []
    for item in result_ds:
        result.append(item)
    joblib.dump(result, Path(__file__).parent /
                f"attributes_clip_tmp_folder/data_{global_rank}.json")
    dist.barrier()

    if is_main_process():
        result = []
        dict_tags = {}
        for rank in range(num_tasks):
            result_file = Path(__file__).parent / \
                f"attributes_clip_tmp_folder/data_{rank}.json"
            res = joblib.load(result_file)
            result += res

        for res in result:
            dict_tags[res[args.id_col]] = res['tags']

        try:
            ds = load_dataset(args.dataset_name,
                              split=args.dataset_split)
        except:
            ds = load_from_disk(args.dataset_name)
        visionencoder_name = DICT_CLIP_NAME[args.vision_encoder_name]
        datasetvocab_name = args.dataset_vocab.split('/')[1].replace('-', '_')
        ds = ds.map(lambda example: {f"Attributes_{visionencoder_name}_{datasetvocab_name}_{args.split_vocab}": dict_tags[example[args.id_col]]},
                    batched=False)
        if args.save_disk:
            ds.save_to_disk(f"{args.dataset_name}_tags")
            # remove
            shutil.rmtree(args.dataset_name)
            # rename file
            os.rename(f"{args.dataset_name}_tags", args.dataset_name)
        else:
            ds.push_to_hub(args.dataset_name)


if __name__ == "__main__":
    ddp_setup()
    main()
    destroy_process_group()



from torchvision import transforms
from PIL import Image
import torch
from datasets import load_dataset, load_from_disk, Dataset
from argparse import ArgumentParser, ArgumentTypeError
import torch.distributed as dist
import itertools
import shutil
import os
from tqdm import tqdm
from torch.distributed import init_process_group, destroy_process_group
import datetime
from transformers import AutoProcessor, BlipForConditionalGeneration
import joblib
from lens.utils import create_sampler, ddp_setup, is_main_process, create_dataloader, default_device
from pathlib import Path


def get_caption_model(model_name, load_8bit=False, device=default_device):
    if load_8bit:
        model = BlipForConditionalGeneration.from_pretrained(model_name,
                                                             torch_dtype=torch.float32 if device=="cpu" else torch.float16,
                                                             device_map={
                                                                 '': device},
                                                             load_in_8bit=True)
    else:
        model = BlipForConditionalGeneration.from_pretrained(model_name,
                                                             torch_dtype=torch.float32 if device=="cpu" else torch.float16)
        model = model.eval()
        model = model.to(device)
    return model


def get_column_name(args):
    column_name = f"blip_caption_{str(args.sample)}_beams_{args.num_beams}_{args.model_name.replace('/','_').replace('-','_')}_max_length_{args.max_length}_hf"
    if args.prefix_text != 'a picture of':
        column_name = f"{column_name}_{args.prefix_text}"
    return column_name


class Captioner:
    def __init__(self,
                 model_name="Salesforce/blip-image-captioning-large",
                 load_8bit=False,
                 device=default_device):
        self.model = get_caption_model(
            model_name, load_8bit=load_8bit, device=device)
        self.transform = AutoProcessor.from_pretrained(model_name)
        self.device = device

    def get_captions(self,
                     ds,
                     sample=False,
                     num_beams=5,
                     top_p=1,
                     max_length=30,
                     min_length=10,
                     batch_size=8,
                     num_workers=0,
                     id_col="id",
                     image_col="image",
                     prefix_text="a picture of",
                     distributed=False):

        dataset = ImageDataset(
            ds, self.transform, id_col=id_col, image_col=image_col, prefix_text=prefix_text)
        sampler = create_sampler(dataset, distributed=distributed)
        dataloader = create_dataloader(
            dataset, sampler, batch_size=batch_size, num_workers=num_workers)
        result = []
        for batch in tqdm(dataloader):
            with torch.no_grad():
                pixel_values = batch['pixel_values'].to(
                    self.device, self.model.dtype)
                input_ids = batch['input_ids'].to(self.device)
                image_id = batch['id']
                len_text = batch['len_text_input']
                captions_ids = self.model.generate(pixel_values=pixel_values,
                                                   input_ids=input_ids,
                                                   do_sample=sample,
                                                   num_beams=num_beams,
                                                   top_p=top_p,
                                                   max_length=max_length,
                                                   min_length=min_length)

            captions = self.transform.batch_decode(
                captions_ids, skip_special_tokens=True)

            for caption, img_id, len_ in zip(captions, image_id, len_text):
                result.append({id_col: img_id.item(),
                               "caption": caption[len_:].strip()})
            del pixel_values, captions_ids
        result_ds = Dataset.from_dict(
            {key: [d[key] for d in result] for key in result[0]})
        return result_ds

    def get_intensive_captions(self,
                               ds,
                               batch_size=8,
                               num_workers=0,
                               id_col="id",
                               image_col="image",
                               prefix_text="a picture of",
                               top_k=50,
                               num_captions=100,
                               distributed=False):

        dataset = ImageDataset(
            ds, self.transform, id_col=id_col, image_col=image_col, prefix_text=prefix_text)
        sampler = create_sampler(dataset, distributed=distributed)
        dataloader = create_dataloader(
            dataset, sampler, batch_size=batch_size, num_workers=num_workers)
        result = []

        for batch in tqdm(dataloader):
            with torch.no_grad():
                pixel_values = batch['pixel_values'].to(
                    self.device, self.model.dtype)
                input_ids = batch['input_ids'].to(self.device)
                attention_mask = batch['attention_mask'].to(self.device)
                image_ids = batch['id']
                caption_ids = self.model.generate(pixel_values=pixel_values,
                                                  input_ids=input_ids,
                                                  attention_mask=attention_mask,
                                                  max_length=30,
                                                  min_length=0,
                                                  do_sample=True,
                                                  top_p=1,
                                                  top_k=top_k,
                                                  num_return_sequences=num_captions,
                                                  repetition_penalty=1)

            captions_text = self.transform.batch_decode(
                caption_ids, skip_special_tokens=True)
            captions_text = [caption[12:].strip() for caption in captions_text]
            captions_text = [captions_text[i:i+num_captions]
                             for i in range(0, len(captions_text), num_captions)]

            for caption, img_id in zip(captions_text, image_ids):
                result.append({id_col: img_id.item(), "caption": caption})
            del pixel_values, input_ids, attention_mask, caption_ids

        result_ds = Dataset.from_dict(
            {key: [d[key] for d in result] for key in result[0]})
        return result_ds


class ImageDataset:
    def __init__(self,
                 ds,
                 transform,
                 id_col="id",
                 image_col="image",
                 prefix_text="a picture of",
                 ):

        self.ds = ds
        self.transform = transform
        self.prefix_text = prefix_text
        self.id_col = id_col
        self.image_col = image_col

    def __getitem__(self, idx):
        row = self.ds[idx]
        image, id = row[self.image_col].convert('RGB'), row[self.id_col]
        len_text_input = len(self.prefix_text)
        inputs = self.transform(image,
                                text=self.prefix_text,
                                return_tensors="pt")
        pixel_values = inputs['pixel_values'].squeeze(0)
        input_ids = inputs['input_ids'].squeeze(0)
        attention_mask = inputs['attention_mask'].squeeze(0)
        return {'pixel_values': pixel_values,
                'input_ids': input_ids,
                'len_text_input': len_text_input,
                'attention_mask': attention_mask,
                'id': torch.tensor(id, dtype=torch.int32)}

    def __len__(self):
        return len(self.ds)


def parse_args():
    args = ArgumentParser(description="")
    # Dataset variables
    args.add_argument('--dataset_name',
                      required=True,
                      type=str,
                      default='Multimodal-Fatima/OxfordPets')

    args.add_argument('--dataset_split',
                      required=True,
                      type=str,
                      default='test')
    args.add_argument('--id_col', type=str, default='id')
    args.add_argument('--image_col', type=str, default='image')
    args.add_argument('--save_disk', action="store_true")

    # Model variables
    args.add_argument('--model_name', required=True, type=str,
                      default='Salesforce/blip2-opt-6.7b')
    args.add_argument('--load_8bit', action="store_true")
    args.add_argument('--sample', action="store_true")
    args.add_argument('--prefix_text', type=str, default='a picture of')
    args.add_argument('--num_beams', type=int, default=5)
    args.add_argument('--top_p', type=float, default=1)
    args.add_argument('--max_length', type=int, default=30)
    args.add_argument('--min_length', type=int, default=10)

    # Parameters
    args.add_argument('--batch_size', type=int, default=8)
    args.add_argument('--num_workers', type=int, default=0)

    # Whether to do intensive or regular captioning
    args.add_argument('--intensive', action="store_true")
    args.add_argument('--top_k', type=int, default=50)
    args.add_argument('--num_captions', type=int, default=100)

    args = args.parse_args()
    return args


def main():
    num_tasks = int(os.environ["WORLD_SIZE"])
    global_rank = int(os.environ["RANK"])
    local_rank = int(os.environ["LOCAL_RANK"])
    args = parse_args()
    ds = load_dataset(args.dataset_name, split=args.dataset_split)
    captioner = Captioner(
        args.model_name, load_8bit=args.load_8bit, device=local_rank)
    if args.intensive:
        result_ds = captioner.get_intensive_captions(ds,
                                                     batch_size=args.batch_size,
                                                     num_workers=args.num_workers,
                                                     id_col=args.id_col,
                                                     image_col=args.image_col,
                                                     prefix_text=args.prefix_text,
                                                     top_k=args.top_k,
                                                     num_captions=args.num_captions,
                                                     distributed=True)
    else:
        result_ds = captioner.get_captions(ds,
                                           sample=args.sample,
                                           num_beams=args.num_beams,
                                           top_p=args.top_p,
                                           max_length=args.max_length,
                                           min_length=args.min_length,
                                           batch_size=args.batch_size,
                                           num_workers=args.num_workers,
                                           id_col=args.id_col,
                                           image_col=args.image_col,
                                           prefix_text=args.prefix_text,
                                           distributed=True)

    os.makedirs(Path(__file__).parent /
                "blip_caption_tmp_folder/", exist_ok=True)
    result = []
    for item in result_ds:
        result.append(item)
    joblib.dump(result, Path(__file__).parent /
                f"blip_caption_tmp_folder/data_{global_rank}.json")
    dist.barrier()

    if is_main_process():
        result = []
        dict_captions = {}

        for rank in range(num_tasks):
            result_file = Path(__file__).parent / \
                f"blip_caption_tmp_folder/data_{rank}.json"
            res = joblib.load(result_file)
            result += res

        for res in result:
            dict_captions[res[args.id_col]] = res['caption']

        # Create dataset
        try:
            ds = load_dataset(args.dataset_name,
                              split=args.dataset_split)
        except:
            ds = load_from_disk(args.dataset_name)

        ds = ds.map(lambda example: {get_column_name(
            args): dict_captions[example[args.id_col]]})

        if args.save_disk:
            ds.save_to_disk(f"{args.dataset_name}_captions")
            # remove
            shutil.rmtree(args.dataset_name)
            # rename file
            os.rename(f"{args.dataset_name}_captions", args.dataset_name)
        else:
            ds.push_to_hub(args.dataset_name)


if __name__ == "__main__":
    ddp_setup()
    main()
    destroy_process_group()



from setuptools import find_packages, setup

INSTALL_REQUIRES = [
    "torch",
    "transformers",
    "evaluate",
    "datasets",
    "tqdm",
    "accelerate",
    "bitsandbytes",
    "ftfy",
    "regex",
    "omegaconf",
    "openai",
    "opencv-python",
    "nltk",
    "scikit-learn",
    "open_clip_torch",
    "wandb",
    "fairscale",
    "huggingface_hub",
]

setup(
    name="llm-lens",
    version="0.0.0.2",
    description=(
        "llm-lens is a Python package for CV as NLP, "
        "where you can run very descriptive image modules "
        "on images, and then pass those descriptions to a "
        "Large Language Model (LLM) to reason about those "
        "images."
    ),
    author="The Contextual AI Tech Team",
    author_email="tech@contextual.ai",
    url="https://github.com/ContextualAI/lens",
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    python_requires=">=3.9",
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



from torch.distributed import init_process_group
import torch
import os
import datetime


default_device = "cuda" if torch.cuda.is_available() else "cpu"


def ddp_setup(): 
    init_process_group(backend="nccl",timeout=datetime.timedelta(seconds=180000))
    torch.cuda.set_device(int(os.environ["LOCAL_RANK"]))


def create_sampler(dataset, distributed=False):
    if distributed:
        sampler = torch.utils.data.DistributedSampler(dataset,shuffle=False)
    else:
        sampler = torch.utils.data.SequentialSampler(dataset)
    return sampler 


def create_dataloader(dataset,sampler, batch_size=8, num_workers=0):
    loader = torch.utils.data.DataLoader(
            dataset,
            batch_size=batch_size,
            num_workers= num_workers,
            pin_memory=True,
            sampler = sampler,
            shuffle= False,
            drop_last= False,
        )              
    return loader


def is_main_process():
    if int(os.environ["RANK"])==0:
        return True
    else:
        return False
